use crate::app_config::AppConfig;
use crate::constants;
use crate::libvirt_lib_structures::{
    DevicesXML, DiskBootXML, DiskDriverXML, DiskReadOnlyXML, DiskSourceXML, DiskTargetXML, DiskXML,
    DomainMemoryXML, DomainXML, DomainXMLUuid, FeaturesXML, GraphicsXML, OSLoaderXML, OSTypeXML,
    ACPIXML, OSXML,
};
use crate::libvirt_rest_structures::LibVirtStructError::StructureExtraction;
use crate::utils::disks_utils::Disk;
use crate::utils::files_utils;
use lazy_regex::regex;
use std::ops::{Div, Mul};

#[derive(thiserror::Error, Debug)]
enum LibVirtStructError {
    #[error("StructureExtractionError: {0}")]
    StructureExtraction(&'static str),
    #[error("DomainExtractionError: {0}")]
    DomainExtraction(String),
    #[error("MBConvertError: {0}")]
    MBConvert(String),
}

#[derive(serde::Serialize)]
pub struct HypervisorInfo {
    pub r#type: String,
    pub hyp_version: u32,
    pub lib_version: u32,
    pub capabilities: String,
    pub free_memory: u64,
    pub hostname: String,
    pub node: HypervisorNodeInfo,
}

#[derive(serde::Serialize)]
pub struct HypervisorNodeInfo {
    pub cpu_model: String,
    /// Memory size in kilobytes
    pub memory_size: u64,
    pub number_of_active_cpus: u32,
    pub cpu_frequency_mhz: u32,
    pub number_of_numa_cell: u32,
    pub number_of_cpu_socket_per_node: u32,
    pub number_of_core_per_sockets: u32,
    pub number_of_threads_per_core: u32,
}

#[derive(serde::Serialize, serde::Deserialize)]
pub enum BootType {
    UEFI,
    UEFISecureBoot,
}

#[derive(serde::Serialize, serde::Deserialize)]
pub enum VMArchitecture {
    #[serde(rename = "i686")]
    I686,
    #[serde(rename = "x86_64")]
    X86_64,
}

#[derive(serde::Serialize, serde::Deserialize)]
pub struct VMInfo {
    /// VM name (alphanumeric characters only)
    pub name: String,
    pub uuid: Option<DomainXMLUuid>,
    pub genid: Option<DomainXMLUuid>,
    pub title: Option<String>,
    pub description: Option<String>,
    pub boot_type: BootType,
    pub architecture: VMArchitecture,
    /// VM allocated memory, in megabytes
    pub memory: usize,
    /// Enable VNC access through admin console
    pub vnc_access: bool,
    /// Attach an ISO file
    pub iso_file: Option<String>,
    /// Storage - https://access.redhat.com/documentation/fr-fr/red_hat_enterprise_linux/6/html/virtualization_administration_guide/sect-virtualization-virtualized_block_devices-adding_storage_devices_to_guests#sect-Virtualization-Adding_storage_devices_to_guests-Adding_file_based_storage_to_a_guest
    pub disks: Vec<Disk>,
    // TODO : autostart
    // TODO : network interface
}

impl VMInfo {
    /// Turn this VM into a domain
    pub fn to_domain(self) -> anyhow::Result<DomainXML> {
        if !regex!("^[a-zA-Z0-9]+$").is_match(&self.name) {
            return Err(StructureExtraction("VM name is invalid!").into());
        }

        let uuid = if let Some(n) = self.uuid {
            if !n.is_valid() {
                return Err(StructureExtraction("VM UUID is invalid!").into());
            }
            n
        } else {
            DomainXMLUuid::new_random()
        };

        if let Some(n) = &self.genid {
            if !n.is_valid() {
                return Err(StructureExtraction("VM genid is invalid!").into());
            }
        }

        if let Some(n) = &self.title {
            if n.contains('\n') {
                return Err(StructureExtraction("VM title contain newline char!").into());
            }
        }

        if self.memory < constants::MIN_VM_MEMORY || self.memory > constants::MAX_VM_MEMORY {
            return Err(StructureExtraction("VM memory is invalid!").into());
        }

        let mut disks = vec![];

        if let Some(iso_file) = &self.iso_file {
            if !files_utils::check_file_name(iso_file) {
                return Err(StructureExtraction("ISO filename is invalid!").into());
            }

            let path = AppConfig::get().iso_storage_path().join(iso_file);

            if !path.exists() {
                return Err(StructureExtraction("Specified ISO file does not exists!").into());
            }

            disks.push(DiskXML {
                r#type: "file".to_string(),
                device: "cdrom".to_string(),
                driver: DiskDriverXML {
                    name: "qemu".to_string(),
                    r#type: "raw".to_string(),
                    cache: "none".to_string(),
                },
                source: DiskSourceXML {
                    file: path.to_string_lossy().to_string(),
                },
                target: DiskTargetXML {
                    dev: "hdc".to_string(),
                    bus: "usb".to_string(),
                },
                readonly: Some(DiskReadOnlyXML {}),
                boot: DiskBootXML {
                    order: "1".to_string(),
                },
                address: None,
            })
        }

        let vnc_graphics = match self.vnc_access {
            true => Some(GraphicsXML {
                r#type: "vnc".to_string(),
                socket: AppConfig::get()
                    .vnc_sockets_path()
                    .join(format!("vnc-{}", self.name))
                    .to_string_lossy()
                    .to_string(),
            }),
            false => None,
        };

        // Check disks name for duplicates
        for disk in &self.disks {
            if self.disks.iter().filter(|d| d.name == disk.name).count() > 1 {
                return Err(StructureExtraction("Two differents disks have the same name!").into());
            }
        }

        // Apply disks configuration
        for disk in self.disks {
            disk.check_config()?;
            disk.apply_config(uuid)?;

            if disk.delete {
                continue;
            }

            disks.push(DiskXML {
                r#type: "file".to_string(),
                device: "disk".to_string(),
                driver: DiskDriverXML {
                    name: "qemu".to_string(),
                    r#type: "raw".to_string(),
                    cache: "none".to_string(),
                },
                source: DiskSourceXML {
                    file: disk.disk_path(uuid).to_string_lossy().to_string(),
                },
                target: DiskTargetXML {
                    dev: format!(
                        "vd{}",
                        ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l"][disks.len()]
                    ),
                    bus: "virtio".to_string(),
                },
                readonly: None,
                boot: DiskBootXML {
                    order: (disks.len() + 1).to_string(),
                },
                address: None,
            })
        }

        Ok(DomainXML {
            r#type: "kvm".to_string(),
            name: self.name,
            uuid: Some(uuid),
            genid: self.genid.map(|i| i.0),
            title: self.title,
            description: self.description,

            os: OSXML {
                r#type: OSTypeXML {
                    arch: match self.architecture {
                        VMArchitecture::I686 => "i686",
                        VMArchitecture::X86_64 => "x86_64",
                    }
                    .to_string(),
                    machine: "q35".to_string(),
                    body: "hvm".to_string(),
                },
                firmware: "efi".to_string(),
                loader: Some(OSLoaderXML {
                    secure: match self.boot_type {
                        BootType::UEFI => "no".to_string(),
                        BootType::UEFISecureBoot => "yes".to_string(),
                    },
                }),
            },

            features: FeaturesXML { acpi: ACPIXML {} },

            devices: DevicesXML {
                graphics: vnc_graphics,
                disks,
            },

            memory: DomainMemoryXML {
                unit: "MB".to_string(),
                memory: self.memory,
            },

            on_poweroff: "preserve".to_string(),
            on_reboot: "restart".to_string(),
            on_crash: "preserve".to_string(),
        })
    }

    /// Turn a domain into a vm
    pub fn from_domain(domain: DomainXML) -> anyhow::Result<Self> {
        Ok(Self {
            name: domain.name,
            uuid: domain.uuid,
            genid: domain.genid.map(DomainXMLUuid),
            title: domain.title,
            description: domain.description,
            boot_type: match domain.os.loader {
                None => BootType::UEFI,
                Some(l) => match l.secure.as_str() {
                    "yes" => BootType::UEFISecureBoot,
                    _ => BootType::UEFI,
                },
            },
            architecture: match domain.os.r#type.arch.as_str() {
                "i686" => VMArchitecture::I686,
                "x86_64" => VMArchitecture::X86_64,
                a => {
                    return Err(LibVirtStructError::DomainExtraction(format!(
                        "Unknown architecture: {a}! "
                    ))
                    .into());
                }
            },
            memory: convert_to_mb(&domain.memory.unit, domain.memory.memory)?,
            vnc_access: domain.devices.graphics.is_some(),
            iso_file: domain
                .devices
                .disks
                .iter()
                .find(|d| d.device == "cdrom")
                .map(|d| d.source.file.rsplit_once('/').unwrap().1.to_string()),

            disks: domain
                .devices
                .disks
                .iter()
                .filter(|d| d.device == "disk")
                .map(|d| Disk::load_from_file(&d.source.file).unwrap())
                .collect(),
        })
    }
}

/// Convert unit to MB
fn convert_to_mb(unit: &str, value: usize) -> anyhow::Result<usize> {
    let fact = match unit {
        "bytes" | "b" => 1f64,
        "KB" => 1000f64,
        "MB" => 1000f64 * 1000f64,
        "GB" => 1000f64 * 1000f64 * 1000f64,
        "TB" => 1000f64 * 1000f64 * 1000f64 * 1000f64,

        "k" | "KiB" => 1024f64,
        "M" | "MiB" => 1024f64 * 1024f64,
        "G" | "GiB" => 1024f64 * 1024f64 * 1024f64,
        "T" | "TiB" => 1024f64 * 1024f64 * 1024f64 * 1024f64,

        _ => {
            return Err(LibVirtStructError::MBConvert(format!("Unknown size unit: {unit}")).into());
        }
    };

    Ok((value as f64).mul(fact.div((1000 * 1000) as f64)).ceil() as usize)
}

#[cfg(test)]
mod test {
    use crate::libvirt_rest_structures::convert_to_mb;

    #[test]
    fn convert_units_mb() {
        assert_eq!(convert_to_mb("MB", 1).unwrap(), 1);
        assert_eq!(convert_to_mb("MB", 1000).unwrap(), 1000);
        assert_eq!(convert_to_mb("GB", 1000).unwrap(), 1000 * 1000);
        assert_eq!(convert_to_mb("GB", 1).unwrap(), 1000);
        assert_eq!(convert_to_mb("GiB", 3).unwrap(), 3222);
        assert_eq!(convert_to_mb("KiB", 488281).unwrap(), 500);
    }
}
